#include "AES256_SIMD.h"
#include <memory.h>

//Cipher(byte in[4*Nb], byte out[4*Nb], word w[Nb*(Nr+1)])
//begin
void AES256_SIMD::Cipher(const BYTE in[4 * Nb], BYTE out[4 * Nb])
{
	//	ASSERT(!decrypting);

	//state = in
	__m128i state = _mm_loadu_si128(reinterpret_cast<const __m128i *>(in));

	//AddRoundKey(state, w[0, Nb-1])
	// xorZ邾
	state = _mm_xor_si128(state, w.w128[0]);

	//for round = 1 step 1 to Nr-1
		//SubBytes(state)
		//ShiftRows(state)
		//MixColumns(state)
		//AddRoundKey(state, w[round*Nb, (round+1)*Nb-1])
	//end for
	state = _mm_aesenc_si128(state, w.w128[1]);
	state = _mm_aesenc_si128(state, w.w128[2]);
	state = _mm_aesenc_si128(state, w.w128[3]);
	state = _mm_aesenc_si128(state, w.w128[4]);
	state = _mm_aesenc_si128(state, w.w128[5]);
	state = _mm_aesenc_si128(state, w.w128[6]);
	state = _mm_aesenc_si128(state, w.w128[7]);
	state = _mm_aesenc_si128(state, w.w128[8]);
	state = _mm_aesenc_si128(state, w.w128[9]);
	state = _mm_aesenc_si128(state, w.w128[10]);
	state = _mm_aesenc_si128(state, w.w128[11]);
	state = _mm_aesenc_si128(state, w.w128[12]);
	state = _mm_aesenc_si128(state, w.w128[13]);

	// Ō̃Eh
	//SubBytes(state)
	//ShiftRows(state)
	//AddRoundKey(state, w[Nr*Nb, (Nr+1)*Nb-1])
	state = _mm_aesenclast_si128(state, w.w128[Nr]);

	//out = state
	_mm_storeu_si128(reinterpret_cast<__m128i *>(out), state);

	//end
}

//EqInvCipher(byte in[4*Nb], byte out[4*Nb], word dw[Nb*(Nr+1)])
//begin
void AES256_SIMD::EqInvCipher(const BYTE in[4 * Nb], BYTE out[4 * Nb])
{
	//	ASSERT(decrypting);

	//byte  state[4,Nb]
	//state = in
	__m128i state = _mm_loadu_si128(reinterpret_cast<const __m128i *>(in));

	//AddRoundKey(state, dw[Nr*Nb, (Nr+1)*Nb-1])
	state = _mm_xor_si128(state, dw.w128[Nr]);

	//for round = Nr-1 step -1 downto 1
		//InvSubBytes(state)
		//InvShiftRows(state)
		//InvMixColumns(state)
		//AddRoundKey(state, dw[round*Nb, (round+1)*Nb-1])
	//end for
	state = _mm_aesdec_si128(state, dw.w128[13]);
	state = _mm_aesdec_si128(state, dw.w128[12]);
	state = _mm_aesdec_si128(state, dw.w128[11]);
	state = _mm_aesdec_si128(state, dw.w128[10]);
	state = _mm_aesdec_si128(state, dw.w128[9]);
	state = _mm_aesdec_si128(state, dw.w128[8]);
	state = _mm_aesdec_si128(state, dw.w128[7]);
	state = _mm_aesdec_si128(state, dw.w128[6]);
	state = _mm_aesdec_si128(state, dw.w128[5]);
	state = _mm_aesdec_si128(state, dw.w128[4]);
	state = _mm_aesdec_si128(state, dw.w128[3]);
	state = _mm_aesdec_si128(state, dw.w128[2]);
	state = _mm_aesdec_si128(state, dw.w128[1]);

	//InvSubBytes(state)
	//InvShiftRows(state)
	//AddRoundKey(state, dw[0, Nb-1])
	state = _mm_aesdeclast_si128(state, dw.w128[0]);

	//out = state
	_mm_storeu_si128(reinterpret_cast<__m128i *>(out), state);
	//end
}

//static const BYTE Rcon[] = {
//	0x8d, 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80, 0x1b, 0x36, 0x6c, 0xd8, 0xab, 0x4d, 0x9a,
//};

//KeyExpansion(byte key[4*Nk], word w[Nb*(Nr+1)], Nk)
//begin
void AES256_SIMD::KeyExpansion(const BYTE key[4 * Nk])
{
	//word  temp
	//i = 0

	//while (i < Nk)
		//w[i] = word(key[4*i], key[4*i+1], key[4*i+2], key[4*i+3])
		//i = i+1
	//end while
	memcpy(w.w32, key, 4 * Nk);

	//i = Nk
	//while (i < Nb * (Nr+1)]
		//temp = w[i-1]
		//if (i mod Nk = 0)
			//temp = SubWord(RotWord(temp)) xor Rcon[i/Nk]
		//else if (Nk > 6 and i mod Nk = 4)
			//temp = SubWord(temp)
		//end if
		//w[i] = w[i-Nk] xor temp
		//i = i + 1
	//end while

//******* i % Nk == 0̂Ƃ̏
//temp = w[i-1]
//temp = SubWord(RotWord(temp)) xor Rcon[i/Nk]
//w[i] = w[i-Nk] xor temp
// w[i-1]܂16oCgE烍[hiw[i-1]ŏDWORDɗj

#define KX_0(i, Rcon) w.w32[i] =				\
	_mm_extract_epi32(							\
		_mm_aeskeygenassist_si128(				\
			*(__m128i*)(&w.w32[i-4]),			\
			Rcon								\
		),										\
		3										\
	) ^ w.w32[i - Nk];

//******* i % Nk == 4̂Ƃ̏
//temp = w[i-1]
//temp = SubWord(temp)
//w[i] = w[i-Nk] xor temp
// w[i-1]܂16oCgE烍[hiw[i-1]ŏDWORDɗj

#define KX_4(i) w.w32[i] =						\
	_mm_extract_epi32(							\
		_mm_aeskeygenassist_si128(				\
			*(__m128i*)(&w.w32[i-4]),			\
			0									\
		),										\
		2										\
	) ^ w.w32[i - Nk];		

//******* ȊÔƂ̏
//temp = w[i-1]
//w[i] = w[i-Nk] xor temp

#define KX(i) w.w32[i] = w.w32[i-1] ^ w.w32[i-Nk];

	// [v̏𗅗

	KX_0(8, 0x01)	KX(9)	KX(10)	KX(11)		KX_4(12)	KX(13)	KX(14)	KX(15)
	KX_0(16, 0x02)	KX(17)	KX(18)	KX(19)		KX_4(20)	KX(21)	KX(22)	KX(23)
	KX_0(24, 0x04)	KX(25)	KX(26)	KX(27)		KX_4(28)	KX(29)	KX(30)	KX(31)
	KX_0(32, 0x08)	KX(33)	KX(34)	KX(35)		KX_4(36)	KX(37)	KX(38)	KX(39)
	KX_0(40, 0x10)	KX(41)	KX(42)	KX(43)		KX_4(44)	KX(45)	KX(46)	KX(47)
	KX_0(48, 0x20)	KX(49)	KX(50)	KX(51)		KX_4(52)	KX(53)	KX(54)	KX(55)
	KX_0(56, 0x40)	KX(57)	KX(58)	KX(59)

	// EqInvCipherp̒ǉ
	if (decrypting) {
		//for i = 0 step 1 to (Nr+1)*Nb-1
			//dw[i] = w[i]
		//end for
		dw.w128[0] = w.w128[0];

		//for round = 1 step 1 to Nr-1
			//InvMixColumns(dw[round*Nb, (round+1)*Nb-1])
		//end for
		dw.w128[1] = _mm_aesimc_si128(w.w128[1]);
		dw.w128[2] = _mm_aesimc_si128(w.w128[2]);
		dw.w128[3] = _mm_aesimc_si128(w.w128[3]);
		dw.w128[4] = _mm_aesimc_si128(w.w128[4]);
		dw.w128[5] = _mm_aesimc_si128(w.w128[5]);
		dw.w128[6] = _mm_aesimc_si128(w.w128[6]);
		dw.w128[7] = _mm_aesimc_si128(w.w128[7]);
		dw.w128[8] = _mm_aesimc_si128(w.w128[8]);
		dw.w128[9] = _mm_aesimc_si128(w.w128[9]);
		dw.w128[10] = _mm_aesimc_si128(w.w128[10]);
		dw.w128[11] = _mm_aesimc_si128(w.w128[11]);
		dw.w128[12] = _mm_aesimc_si128(w.w128[12]);
		dw.w128[13] = _mm_aesimc_si128(w.w128[13]);

		dw.w128[Nr] = w.w128[Nr];
	}
//end
}

